// Gray - A simple ray tracing program
// Copyright (C) 2008-2018 Eric Bechet
//
// See the COPYING file for contributions and license information.
// Please report all bugs and problems to <bechet@cadxfem.org>.
//

#include "main.h"
#include "glframebuffer.h"
#include "scene.h"
#include "materials.h"
#include "meshmodel.h"
#include <iostream>
#include <algorithm>
#include <vector>
#include "QuasiMonteCarlo.h"
#ifdef GRAY_HAVE_GNURBS
#include "gnurbs-interface.h"
#endif
#include "Texture.h"

int main ( int argc, char *argv[] )
{
#ifdef GRAY_HAVE_GNURBS
  test();
#endif

  Gscene sc;
  sc.setOversampling ( 4 );
  sc.setMaxRecurLevel ( 20 );

  //Textures loading
  Texture tex1 ( "textures/checker.bmp" );
  Texture tex2 ( "textures/ground.png" );
  Texture tex3 ( "textures/checkergreen.png" );
  Texture tex4 ( "textures/checker_normal.png" );
  Texture tex5 ( "textures/checkergreen_normal.png" );
  Texture tex6 ( "textures/ground_normal.png" );

  // Various materials
  GmaterialTransparent mat0 ( Color ( 0,0,0 ), Color ( 0,0,0 ), Color ( 0.9931,0.9931,1 ), 100, 4, 1.52 );
  /*  GmaterialLambertBlinnPhong mat1(Color(1.0,1.0,1.0), Color(0.3,0.3,0.3), 100., 4.0);
    GmaterialLambertBlinnPhong mat2(Color(1.0,1.0,1.0), Color(0.3,0.3,0.3), 100., 4.0);
    GmaterialLambertBlinnPhong mat3(Color(1.0,1.0,1.0), Color(0.3,0.3,0.3), 100., 4.0);
    GmaterialLambertBlinnPhong mat4(Color(1.0,1.0,1.0), Color(0.3,0.3,0.3), 100., 4.0);
    GmaterialLambertBlinnPhong mat5(Color(1,1,1), Color(0.1,0.1,0.1), 100., 0);*/
  GmaterialLambertBlinnPhong mat1 ( Color ( 1.0,1.0,0.0 ), Color ( 0.3,0.3,0.3 ), 100., 4.0 );
  GmaterialLambertBlinnPhong mat2 ( Color ( 0.0,1.0,1.0 ), Color ( 0.3,0.3,0.3 ), 100., 4.0 );
  GmaterialLambertBlinnPhong mat3 ( Color ( 1.0,0.0,1.0 ), Color ( 0.3,0.3,0.3 ), 100., 4.0 );
  GmaterialLambertBlinnPhong mat4 ( Color ( 0.0,0.0,0.0 ), Color ( 0.3,0.3,0.3 ), 100., 4.0 );
  GmaterialLambertBlinnPhong mat5 ( Color ( 0.4,0.4,0.6 ), Color ( 0.1,0.1,0.1 ), 100., 0 );




  GmaterialLambertBlinnPhong mat6 ( Color ( 111./255.,49./255.,152./255. ),Color ( 0.3,0.3,0.3 ), 12.,1.0 );
  //Texture support
  mat1.setTexture ( &tex3 );
  mat1.setNormalMap ( &tex5 );
  mat2.setTexture ( &tex1 );
  mat2.setNormalMap ( &tex4 );
  mat3.setTexture ( &tex1 );
  mat3.setNormalMap ( &tex4 );
  mat4.setTexture ( &tex3 );
  mat4.setNormalMap ( &tex5 );
  mat5.setTexture ( &tex2, 80.0, 80.0 );
  mat5.setNormalMap ( &tex6 );

  Gmeshmodel *m=new Gmeshmodel ( &mat0 );
  m->linearInterpolation = true;
  m->read_ply ( "ply/monkey",60,0,40,0 ); // 505 vertices and 500 faces
  m->mergeCloseVertices();
  m->subdivideCatmulClark ( 2 );
  m->compute_normals();
  sc.insertModel ( m );

  // group of Gsphere and Gdisc objects
  sc.insertModel ( new Gsphere ( 30.0,Point ( 100,30,0 ),&mat1 ) );
  sc.insertModel ( new Gsphere ( 30.0,Point ( -100,30,0 ),&mat2 ) );
  sc.insertModel ( new Gsphere ( 30.0,Point ( 0,30,100 ),&mat3 ) );
  sc.insertModel ( new Gsphere ( 30.0,Point ( 0,30,-100 ),&mat4 ) );

  sc.insertModel ( new Gdisc ( 10000,Point ( 0,0,0 ), Point ( 0,1,0 ), &mat5 ) );
//  sc.insertModel(new Gsphere(50.0,Point (0,50,-300),&mat0));

  // Lights
  Glight *light1 = new Glight ( Point ( -200,300,-200 ), 1, Color ( 1.0,1.0,1.0 ) );
  sc.insertLight ( light1 );
  sc.setTypicalLightDist ( light1->getPosition().norm() );

  Glight *light2 = new Glight ( Point ( 200,300,200 ), 1, Color ( 1.0,1.0,1.0 ) );
  sc.insertLight ( light2 );
  sc.setTypicalLightDist ( light2->getPosition().norm() );

  // Viewpoint
  sc.setViewpoint ( Point ( 150,120,100 ) );
  //sc.setViewpoint(Point (0,120,-500));
  //sc.setViewpoint(light1->getPosition());

  // Camera
  Camera *cam = new Camera();
  sc.setCam ( cam );
  cam->setFocusDist ( 22.36 );
  cam->setDirection ( Point ( 0,40,0 ) );

  light1->setCache ( new ConicCache ( cam->getDirection()-light1->getPosition(), 0.9925 ) );
  light2->setCache ( new ConicCache ( cam->getDirection()-light2->getPosition(), 0.9925 ) );

#ifdef GRAY_HAVE_PHOTONMAP
  time_t start = time ( 0 );
  std::cout << "Initializing photon mapping. " << std::endl;
  unsigned long nbPhotons = sc.initPhotonMap ( 5E5, 1, 1E6 );
  std::cout << "Done in " << time ( 0 ) - start << " seconds. " << nbPhotons << " photons generated. " << std::endl;
#endif

  sc.setDirectLighting ( true );
#ifdef GRAY_HAVE_PHOTONMAP
  sc.setPhotonLighting ( true );
#endif

  // Initialisation du frameBuffer (image raytracée) et de la fenêtre.
  gl_framebuffer *fb= new gl_framebuffer ( 500, 500,&sc );
  fb->display();
  
  delete cam;
  delete fb;
  return 0;
}
// kate: indent-mode cstyle; indent-width 2; replace-tabs on; 
